import torch
import torch.nn as nn

from typing import List
from .allinone import AllInOne
from ..typings import AllInOneOutput


class Ensemble(nn.Module):
  def __init__(self, models: List[AllInOne]):
    super().__init__()

    cfg = models[0].cfg.copy()
    cfg.best_threshold_beat = sum([model.cfg.best_threshold_beat for model in models]) / len(models)
    cfg.best_threshold_downbeat = sum([model.cfg.best_threshold_downbeat for model in models]) / len(models)

    self.cfg = cfg
    self.models = models

  def forward(self, x):
    outputs: List[AllInOneOutput] = [model(x) for model in self.models]
    avg = AllInOneOutput(
      logits_beat=torch.stack([output.logits_beat for output in outputs], dim=0).mean(dim=0),
      logits_downbeat=torch.stack([output.logits_downbeat for output in outputs], dim=0).mean(dim=0),
      logits_section=torch.stack([output.logits_section for output in outputs], dim=0).mean(dim=0),
      logits_function=torch.stack([output.logits_function for output in outputs], dim=0).mean(dim=0),
      embeddings=torch.stack([output.embeddings for output in outputs], dim=-1),
    )

    return avg



import re

from pathlib import Path
from .typings import PathLike


def compact_json_number_array(json_str: str):
  """Compact numbers (including floats) in JSON arrays to be on the same line."""
  return re.sub(
    r'(\[\n(?:\s*\d+(\.\d+)?,\n)+\s*\d+(\.\d+)?\n\s*\])',
    lambda m: m.group(1).replace('\n', '').replace(' ', ''),
    json_str
  )


def mkpath(path: PathLike):
  return Path(path).expanduser().resolve()



import allin1

from pathlib import Path
from allin1.utils import mkpath


def test_visualize():
  result = allin1.analyze(
    paths=Path(__file__).resolve().parent / 'test.mp3',
    keep_byproducts=True,
  )
  allin1.visualize(result)


def test_visualize_save():
  result = allin1.analyze(
    paths=Path(__file__).resolve().parent / 'test.mp3',
    keep_byproducts=True,
  )
  allin1.visualize(
    result,
    out_dir='./plots',
  )
  assert mkpath('./plots/test.pdf').is_file()



from typing import List, Optional, Any
from dataclasses import dataclass, field
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING

HARMONIX_LABELS = [
  'start',
  'end',
  'intro',
  'outro',
  'break',
  'bridge',
  'inst',
  'solo',
  'verse',
  'chorus',
]


@dataclass
class DataConfig:
  name: str

  demixed: bool
  num_instruments: int
  num_labels: int

  path_base_dir: str
  path_track_dir: str
  path_demixed_dir: str
  path_feature_dir: str
  path_no_demixed_feature_dir: str

  duration_min: float
  duration_max: float

  demucs_model: str = 'htdemucs'


@dataclass
class HarmonixConfig(DataConfig):
  name: str = 'harmonix'

  demixed: bool = True
  num_instruments: int = 4
  num_labels: int = 10

  path_base_dir: str = './data/harmonix/'
  path_track_dir: str = './data/harmonix/tracks/'
  path_demixed_dir: str = './data/harmonix/demixed/'
  path_feature_dir: str = './data/harmonix/features/'
  path_no_demixed_feature_dir: str = './data/harmonix/features_no_demixed/'
  path_metadata: str = './data/harmonix/metadata.csv'

  duration_min: int = 76
  duration_max: int = 660


defaults = [
  '_self_',
  {'data': 'harmonix'},
]


@dataclass
class Config:
  debug: bool = False
  sanity_check: bool = False
  sanity_check_size: int = 1
  offline: bool = False

  case: Optional[str] = None
  model: str = 'allinone'  # allin1, tcn

  data: DataConfig = MISSING
  defaults: List[Any] = field(default_factory=lambda: defaults)

  # Data configurations --------------------------------------------------
  sample_rate: int = 44100
  window_size: int = 2048
  num_bands: int = 12
  hop_size: int = 441  # FPS=100
  fps: int = 100
  fmin: int = 30
  fmax: int = 17000
  demucs_model: str = 'htdemucs'

  # Multi-task learning configurations ------------------------------------
  learn_rhythm: bool = True
  learn_structure: bool = True
  learn_segment: bool = True
  learn_label: bool = True

  # Training configurations -----------------------------------------------
  segment_size: Optional[float] = 300
  batch_size: int = 1

  optimizer: str = 'radam'
  sched: Optional[str] = 'plateau'
  lookahead: bool = False

  lr: float = 0.005
  warmup_lr: float = 1e-5
  warmup_epochs: int = 0
  cooldown_epochs: int = 0
  min_lr: float = 1e-7
  max_epochs: int = -1

  # Plateau scheduler.
  decay_rate: float = 0.3
  patience_epochs: int = 5
  eval_metric: str = 'val/loss'
  epochs: int = 10  # not used. just for creating the scheduler

  validation_interval_epochs: int = 3
  early_stopping_patience: int = 10

  weight_decay: float = 0.00025
  swa_lr: float = 0.15
  gradient_clip: float = 0.5

  # Model configurations --------------------------------------------------
  threshold_beat: float = 0.19
  threshold_downbeat: float = 0.19
  threshold_section: float = 0.05

  best_threshold_beat: Optional[float] = None
  best_threshold_downbeat: Optional[float] = None

  instrument_attention: bool = True
  double_attention: bool = True

  depth: int = 11
  dilation_factor: int = 2
  dilation_max: int = 3200  # 32 seconds, not in use
  num_heads: int = 2
  kernel_size: int = 5

  dim_input: int = 81
  dim_embed: int = 24
  mlp_ratio: float = 4.0
  qkv_bias: bool = True

  drop_conv: float = 0.2
  drop_path: float = 0.1
  drop_hidden: float = 0.2
  drop_attention: float = 0.2
  drop_last: float = 0.0

  act_conv: str = 'elu'
  act_transformer: str = 'gelu'

  layer_norm_eps: float = 1e-5

  # Loss configurations ---------------------------------------------------
  loss_weight_beat: float = 1.
  loss_weight_downbeat: float = 3.
  loss_weight_section: float = 15.
  loss_weight_function: float = 0.1

  # Misc ------------------------------------------------------------------
  seed: int = 1234
  fold: int = 2
  aafold: Optional[int] = None
  total_folds: int = 8

  bpm_min: int = 55
  bpm_max: int = 240
  min_hops_per_beat: int = 24  # 60 / max_bpm * sample_rate / hop_size


cs = ConfigStore.instance()
cs.store(name='config', node=Config)
cs.store(group='data', name=HarmonixConfig.name, node=HarmonixConfig)
cs.store(name=HarmonixConfig.name, node=HarmonixConfig)



import argparse
import torch

from pathlib import Path
from .analyze import analyze


def make_parser():
  cwd = Path.cwd()
  parser = argparse.ArgumentParser()
  parser.add_argument('paths', nargs='+', type=Path, default=[], help='Path to tracks')
  parser.add_argument('-a', '--activ', action='store_true',
                      help='Save frame-level raw activations from sigmoid and softmax (default: False)')
  parser.add_argument('-e', '--embed', action='store_true',
                      help='Save frame-level embeddings (default: False)')
  parser.add_argument('-o', '--out-dir', type=Path, default=cwd / './structures',
                      help='Path to a directory to store analysis results (default: ./structures)')
  parser.add_argument('-v', '--visualize', action='store_true',
                      help='Visualize analysis results and save plots (default: False)')
  parser.add_argument('-p', '--plot-dir', type=Path, default=cwd / './plots',
                      help='Path to a directory to store plots (default: ./plots)')
  parser.add_argument('-m', '--model', type=str, default='harmonix-all',
                      help='Name of the pretrained model to use (default: harmonix-all)')
  parser.add_argument('-d', '--device', type=str, default='cuda' if torch.cuda.is_available() else 'cpu',
                      help='Device to use (default: cuda if available else cpu)')
  parser.add_argument('-k', '--keep-byproducts', action='store_true',
                      help='Keep demixed audio files and spectrograms (default: False)')
  parser.add_argument('--demix-dir', type=Path, default=cwd / 'demixed',
                      help='Path to a directory to store demixed tracks (default: ./demixed)')
  parser.add_argument('--spec-dir', type=Path, default=cwd / 'spectrograms',
                      help='Path to a directory to store spectrograms (default: ./spectrograms)')

  return parser


def main():
  parser = make_parser()
  args = parser.parse_args()

  if not args.paths:
    raise ValueError('At least one path must be specified.')

  assert args.out_dir is not None, 'Output directory must be specified with --out-dir'

  analyze(
    paths=args.paths,
    out_dir=args.out_dir,
    plot_dir=args.plot_dir if args.visualize else None,
    model=args.model,
    device=args.device,
    include_activations=args.activ,
    include_embeddings=args.embed,
    demix_dir=args.demix_dir,
    spec_dir=args.spec_dir,
    keep_byproducts=args.keep_byproducts,
  )

  print(f'=> Analysis results are successfully saved to {args.out_dir}')


if __name__ == '__main__':
  main()



import torch

from typing import Optional
from omegaconf import OmegaConf
from huggingface_hub import hf_hub_download
from .allinone import AllInOne
from .ensemble import Ensemble
from ..typings import PathLike

NAME_TO_FILE = {
  'harmonix-fold0': 'harmonix-fold0-0vra4ys2.pth',
  'harmonix-fold1': 'harmonix-fold1-3ozjhtsj.pth',
  'harmonix-fold2': 'harmonix-fold2-gmgo0nsy.pth',
  'harmonix-fold3': 'harmonix-fold3-i92b7m8p.pth',
  'harmonix-fold4': 'harmonix-fold4-1bql5qo0.pth',
  'harmonix-fold5': 'harmonix-fold5-x4z5zeef.pth',
  'harmonix-fold6': 'harmonix-fold6-x7t226rq.pth',
  'harmonix-fold7': 'harmonix-fold7-qwwskhg6.pth',
}

ENSEMBLE_MODELS = {
  'harmonix-all': [
    'harmonix-fold0',
    'harmonix-fold1',
    'harmonix-fold2',
    'harmonix-fold3',
    'harmonix-fold4',
    'harmonix-fold5',
    'harmonix-fold6',
    'harmonix-fold7',
  ],
}


def load_pretrained_model(
  model_name: Optional[str] = None,
  cache_dir: Optional[PathLike] = None,
  device=None,
):
  if model_name in ENSEMBLE_MODELS:
    return load_ensemble_model(model_name, cache_dir, device)

  model_name = model_name or list(NAME_TO_FILE.keys())[0]
  assert model_name in NAME_TO_FILE, f'Unknown model name: {model_name} (expected one of {list(NAME_TO_FILE.keys())})'

  if device is None:
    if torch.cuda.device_count():
      device = 'cuda'
    else:
      device = 'cpu'

  filename = NAME_TO_FILE[model_name]
  checkpoint_path = hf_hub_download(repo_id='taejunkim/allinone', filename=filename, cache_dir=cache_dir)

  checkpoint = torch.load(checkpoint_path, map_location=device)
  config = OmegaConf.create(checkpoint['config'])

  model = AllInOne(config).to(device)
  model.load_state_dict(checkpoint['state_dict'])
  model.eval()

  return model


def load_ensemble_model(
  model_name: Optional[str] = None,
  cache_dir: Optional[PathLike] = None,
  device=None,
):
  models = []
  for model_name in ENSEMBLE_MODELS[model_name]:
    model = load_pretrained_model(model_name, cache_dir, device)
    models.append(model)

  ensemble = Ensemble(models).to(device)
  ensemble.eval()

  return ensemble



import numpy as np
from pathlib import Path
from typing import List, Tuple
from tqdm import tqdm
from multiprocessing import Pool
from madmom.audio.signal import FramedSignalProcessor, Signal
from madmom.audio.stft import ShortTimeFourierTransformProcessor
from madmom.processors import SequentialProcessor
from madmom.audio.spectrogram import FilteredSpectrogramProcessor, LogarithmicSpectrogramProcessor


def extract_spectrograms(demix_paths: List[Path], spec_dir: Path):
  todos = []
  spec_paths = []
  for src in demix_paths:
    dst = spec_dir / f'{src.stem}.npy'
    spec_paths.append(dst)
    if dst.is_file():
      continue
    todos.append((src, dst))

  existing = len(spec_paths) - len(todos)
  print(f'=> Found {existing} spectrograms already extracted, {len(todos)} to extract.')

  if todos:
    # Define a pre-processing chain, which is copied from madmom.
    frames = FramedSignalProcessor(
      frame_size=2048,
      fps=int(44100 / 441)
    )
    stft = ShortTimeFourierTransformProcessor()  # caching FFT window
    filt = FilteredSpectrogramProcessor(
      num_bands=12,
      fmin=30,
      fmax=17000,
      norm_filters=True
    )
    spec = LogarithmicSpectrogramProcessor(mul=1, add=1)
    processor = SequentialProcessor([frames, stft, filt, spec])

    # Process all tracks using multiprocessing.
    with Pool() as pool:
      iterator = pool.imap(_extract_spectrogram, [
        (src, dst, processor)
        for src, dst in todos
      ])
      for _ in tqdm(iterator, total=len(todos), desc='Extracting spectrograms'):
        pass

  return spec_paths


def _extract_spectrogram(args: Tuple[Path, Path, SequentialProcessor]):
  src, dst, processor = args

  dst.parent.mkdir(parents=True, exist_ok=True)

  sig_bass = Signal(src / 'bass.wav', num_channels=1)
  sig_drums = Signal(src / 'drums.wav', num_channels=1)
  sig_other = Signal(src / 'other.wav', num_channels=1)
  sig_vocals = Signal(src / 'vocals.wav', num_channels=1)

  spec_bass = processor(sig_bass)
  spec_drums = processor(sig_drums)
  spec_others = processor(sig_other)
  spec_vocals = processor(sig_vocals)

  spec = np.stack([spec_bass, spec_drums, spec_others, spec_vocals])  # instruments, frames, bins

  np.save(str(dst), spec)



import sys
import subprocess
import torch

from pathlib import Path
from typing import List, Union


def demix(paths: List[Path], demix_dir: Path, device: Union[str, torch.device]):
  """Demixes the audio file into its sources."""
  todos = []
  demix_paths = []
  for path in paths:
    out_dir = demix_dir / 'htdemucs' / path.stem
    demix_paths.append(out_dir)
    if out_dir.is_dir():
      if (
        (out_dir / 'bass.wav').is_file() and
        (out_dir / 'drums.wav').is_file() and
        (out_dir / 'other.wav').is_file() and
        (out_dir / 'vocals.wav').is_file()
      ):
        continue
    todos.append(path)

  existing = len(paths) - len(todos)
  print(f'=> Found {existing} tracks already demixed, {len(todos)} to demix.')

  if todos:
    subprocess.run(
      [
        sys.executable, '-m', 'demucs.separate',
        '--out', demix_dir.as_posix(),
        '--name', 'htdemucs',
        '--device', str(device),
        *[path.as_posix() for path in todos],
      ],
      check=True,
    )

  return demix_paths



import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.gridspec as gridspec
import librosa.feature

from typing import Union, List, Mapping

from .typings import AnalysisResult, PathLike
from .utils import mkpath

HARMONIX_COLORS = {
  'start': 'black',
  'end': 'black',
  'intro': 1,
  'outro': 1,
  'break': 2,
  'bridge': 2,
  'inst': 3,
  'solo': 3,
  'verse': 4,
  'chorus': 5,
}


def visualize(
  results: Union[AnalysisResult, List[AnalysisResult]],
  out_dir: PathLike = None,
):
  return_list = True
  if not isinstance(results, list):
    return_list = False
    results = [results]

  figs = [_plot(r) for r in results]

  # Save figures to out_dir
  if out_dir is not None:
    out_dir = mkpath(out_dir)
    out_dir.mkdir(parents=True, exist_ok=True)
    for result, fig in zip(results, figs):
      fig.savefig(out_dir / f'{result.path.stem}.pdf', bbox_inches='tight')

  # return the first figure if results was not a list
  if not return_list:
    return figs[0]
  return figs


def _plot(
  result: AnalysisResult,
  colors: Mapping[str, int] = None,
  color_map: str = 'viridis',
):
  if colors is None:
    colors = HARMONIX_COLORS

  y, sr = librosa.load(result.path, sr=None, mono=True)
  rms = librosa.feature.rms(y=y, frame_length=4096, hop_length=1024)[0]

  fig = plt.figure(figsize=(12, 2))
  gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1])
  ax0 = plt.subplot(gs[0])
  ax1 = plt.subplot(gs[1])

  ax0.plot(rms, color='black', linewidth=1)
  ax0.set_xlim(0, len(rms) - 1)
  ax0.set_ylim(0, None)
  ax0.set_ylabel('RMS')
  ax0.set_xticks([])

  cmap = plt.get_cmap(color_map)
  max_color = max(c for c in colors.values() if not isinstance(c, str))
  min_color = min(c for c in colors.values() if not isinstance(c, str))
  for segment in result.segments:
    color = colors[segment.label]
    if not isinstance(color, str):
      color = (color - min_color) / (max_color - min_color)
      color = cmap(color)

    ax1.axvspan(segment.start, segment.end, color=color)
    ax1.axvline(segment.start, color='black', linewidth=1)
    if segment.label not in ['start', 'end']:
      ax1.text(
        (segment.end - segment.start) / 2 + segment.start,
        0.5,
        segment.label,
        color=_get_text_color(color),
        fontsize=12,
        weight='bold',
        horizontalalignment='center',
        verticalalignment='center',
      )
  ax1.set_xlim(0, result.segments[-1].end)
  tick_segments = [s for s in result.segments if s.label != 'start']
  ax1.set_xticks(
    [s.start for s in tick_segments],
    [f'{round(s.start // 60)}:{round(s.start % 60):02}' for s in tick_segments],
  )
  ax1.set_xlabel('Time (min:sec)')
  ax1.set_yticks([])

  # set title
  ax0.set_title(result.path.name)
  fig.tight_layout()
  fig.subplots_adjust(hspace=0)

  return fig


def _get_text_color(bg_color):
  bg_color = mcolors.to_rgb(bg_color)
  luminance = 0.299 * bg_color[0] + 0.587 * bg_color[1] + 0.114 * bg_color[2]

  if luminance > 0.5:
    return 'black'
  else:
    return 'white'



import torch
from os import PathLike
from pathlib import Path
from typing import List, Dict, Optional, Union
from dataclasses import dataclass
from numpy.typing import NDArray

PathLike = Union[str, PathLike]


@dataclass
class AllInOneOutput:
  logits_beat: torch.FloatTensor = None
  logits_downbeat: torch.FloatTensor = None
  logits_section: torch.FloatTensor = None
  logits_function: torch.FloatTensor = None
  embeddings: torch.FloatTensor = None


@dataclass
class Segment:
  start: float
  end: float
  label: str


@dataclass
class AnalysisResult:
  path: Path
  bpm: int
  beats: List[float]
  downbeats: List[float]
  beat_positions: List[int]
  segments: List[Segment]
  activations: Optional[Dict[str, NDArray]] = None
  embeddings: Optional[NDArray] = None



import numpy as np
import torch

from typing import List, Union
from tqdm import tqdm
from .demix import demix
from .spectrogram import extract_spectrograms
from .models import load_pretrained_model
from .visualize import visualize
from .postprocessing import (
  postprocess_metrical_structure,
  postprocess_functional_structure,
  estimate_tempo_from_beats,
)
from .helpers import (
  compute_activations,
  expand_paths,
  check_paths,
  rmdir_if_empty,
  save_results,
)
from .utils import mkpath
from .typings import  AnalysisResult, PathLike

def analyze(
  paths: Union[List[PathLike], PathLike],
  out_dir: PathLike = None,
  plot_dir: PathLike = None,
  model: str = 'harmonix-all',
  device: str = 'cuda' if torch.cuda.is_available() else 'cpu',
  include_activations: bool = False,
  include_embeddings: bool = False,
  demix_dir: PathLike = './demixed',
  spec_dir: PathLike = './spectrograms',
  keep_byproducts: bool = False,
):
  return_list = True
  if not isinstance(paths, list):
    return_list = False
    paths = [paths]
  paths = [mkpath(p) for p in paths]
  paths = expand_paths(paths)
  check_paths(paths)
  demix_dir = mkpath(demix_dir)
  spec_dir = mkpath(spec_dir)
  device = torch.device(device)
  print(f'=> Found {len(paths)} tracks to analyze.')

  demix_paths = demix(paths, demix_dir, device)

  spec_paths = extract_spectrograms(demix_paths, spec_dir)

  model = load_pretrained_model(
    model_name=model,
    device=device,
  )

  results = []
  with torch.no_grad():
    pbar = tqdm(zip(paths, spec_paths), total=len(paths))
    for path, spec_path in pbar:
      pbar.set_description(f'Analyzing {path.name}')

      spec = np.load(spec_path)
      spec = torch.from_numpy(spec).unsqueeze(0).to(device)

      logits = model(spec)

      metrical_structure = postprocess_metrical_structure(logits, model.cfg)
      functional_structure = postprocess_functional_structure(logits, model.cfg)
      bpm = estimate_tempo_from_beats(metrical_structure['beats'])

      result = AnalysisResult(
        path=path,
        bpm=bpm,
        segments=functional_structure,
        **metrical_structure,
      )
      results.append(result)

      if include_activations:
        activations = compute_activations(logits)
        result.activations = activations

      if include_embeddings:
        result.embeddings = logits.embeddings[0].cpu().numpy()

  if not keep_byproducts:
    for path in demix_paths:
      for stem in ['bass', 'drums', 'other', 'vocals']:
        (path / f'{stem}.wav').unlink(missing_ok=True)
      rmdir_if_empty(path)
    rmdir_if_empty(demix_dir / 'htdemucs')
    rmdir_if_empty(demix_dir)

    for path in spec_paths:
      path.unlink(missing_ok=True)
    rmdir_if_empty(spec_dir)

  if out_dir is not None:
    save_results(results, out_dir)

  if plot_dir is not None:
    visualize(results, out_dir=plot_dir)
    print(f'=> Plots are successfully saved to {plot_dir}')

  if not return_list:
    return results[0]
  return results




import numpy as np
from typing import List


def estimate_tempo_from_beats(
  beats: List[float],
):
  beats = np.array(beats)
  beat_interval = np.diff(beats)
  bpm = 60. / beat_interval
  bpm = bpm.round().astype(int)
  bincount = np.bincount(bpm)
  bpm_range = np.arange(len(bincount))
  bpm_strength = bincount / bincount.sum()
  bpm_cand = np.stack([bpm_range, bpm_strength], axis=-1)
  bpm_cand = bpm_cand[np.argsort(bpm_strength)[::-1]]
  bpm_cand = bpm_cand[bpm_cand[:, 1] > 0]

  bpm_est = bpm_cand[0, 0]
  bpm_est = int(bpm_est)

  return bpm_est
